using System;
using System.Collections.Generic;
using System.Diagnostics;
using LiteNetLib;
using LiteNetLib.Utils;

using Syntriax.Engine.Core;
using Syntriax.Engine.Network.Abstract;

namespace Syntriax.Engine.Network;

public abstract class NetworkBase : BehaviourOverride, INetworkCommunicator
{
    private readonly NetPacketProcessor netPacketProcessor = new();

    private readonly Dictionary<uint, INetworkEntity> networkEntities = [];

    private BehaviourCollector<INetworkEntity> networkEntityCollector = null!;

    public EventBasedNetListener Listener { get; private set; } = null!;
    public NetManager Manager { get; private set; } = null!;

    public NetworkBase()
    {
        Priority = 10;

        Listener = new EventBasedNetListener();
        Manager = new NetManager(Listener);

        Listener.NetworkReceiveEvent += NetworkReceiveEvent;
        netPacketProcessor.SubscribeReusable<NetworkPacket<int>, NetPeer>(OnPacketArrived);
    }

    private void OnPacketArrived(NetworkPacket<int> packet, NetPeer peer)
    {
        Debug.WriteLine($"Packet Arrived for {packet.NetworkId}: {packet.Data}");
    }

    protected override void OnInitialize()
    {
        base.OnInitialize();

        networkEntityCollector = new(GameObject.GameManager);
        networkEntityCollector.OnCollected += OnNetworkEntityCollected;
        networkEntityCollector.OnRemoved += OnNetworkEntityRemoved;
    }

    protected override void OnFinalize()
    {
        networkEntityCollector.OnCollected -= OnNetworkEntityCollected;
        networkEntityCollector.OnRemoved -= OnNetworkEntityRemoved;
        Stop();
    }

    private void OnNetworkEntityCollected(BehaviourCollector<INetworkEntity> sender, INetworkEntity behaviourCollected)
        => networkEntities.Add(behaviourCollected.NetworkId, behaviourCollected);

    private void OnNetworkEntityRemoved(BehaviourCollector<INetworkEntity> sender, INetworkEntity behaviourRemoved)
        => networkEntities.Remove(behaviourRemoved.NetworkId);

    private void NetworkReceiveEvent(NetPeer peer, NetPacketReader reader, byte channel, DeliveryMethod deliveryMethod)
    {
        netPacketProcessor.ReadAllPackets(reader, peer);
    }

    public void PollEvents() => Manager.PollEvents();
    public void Stop() => Manager.Stop();

    protected override void OnUpdate() => PollEvents();
}
